//******************************************************************************
//Name          : FFT.cs
//Description   : Implementation to experience the beauty and the pitfalls of FFT
//Author        : Dr. Oliver Kluge, 90411 Nuernberg, Germany
//Copyright(C)  : Oliver Kluge 2014 
//------------------------------------------------------------------------------
//******************************************************************************

using System;
using System.Windows.Forms;
using System.Windows.Forms.DataVisualization.Charting;
using Display;


public struct complex
{
    public double re;
    public double im;
}


class Spectrum
{
    const int samplerate = 128;                         // Abtastwerte im Abtastintervall
    const double PI      = 3.141592653589793;

    public double[] signal = new double[samplerate];    // Signalverlauf
    public double[] arg = new double[samplerate];
    public complex[] data = new complex[samplerate];    // komplexe Daten
    public double[] AS   = new double[samplerate];      // Amplitudenspektrum
    public double[] PS = new double[samplerate];        // Phasenspektrum

    public void CreateSignal(int select)
    {
        if (select >= 1 && select <= 8)
        {
            int width = 0;

            // 1/1 Puls
            if (select == 1) width = samplerate;
            // 1/2 Puls
            if (select == 2) width = samplerate/2;
            // 1/4 Puls
            if (select == 3) width = samplerate/4;
            // 1/8 Puls
            if (select == 4) width = samplerate/8;
            // 1/16 Puls
            if (select == 5) width = samplerate/16;
            // 1/32 Puls
            if (select == 6) width = samplerate/32;
            // 1/64 Puls
            if (select == 7) width = samplerate/64;
            // 1/128 Puls
            if (select == 8) width = samplerate/128;

            for (int k = 0; k < samplerate; k++)
            {
                if (k < width)
                {
                    signal[k] = 1;
                }
                else
                {
                    signal[k] = 0;
                }
            }
        }

        // Deltakamm
        if (select == 9)
        {
            for (int k = 0; k < samplerate; k++)
            {
                if (k == 0  || k == 16 || k == 32 || k == 48  ||
                    k == 64 || k == 80 || k == 96 || k == 112 || k == 128)
                {
                    signal[k] = 1;
                }
                else
                {
                    signal[k] = 0;
                }
            }
        }

        // Spaltfunktion
        if (select == 10)
        {
            double f = 16.0;

            for (int k = 0; k < samplerate; k++)
            {
                signal[k] = 15 * Math.Sin(0.001 + PI * f * (k - samplerate / 2) / samplerate) / (0.001 + PI * f * (k - samplerate / 2) / samplerate);
            }
        }

        // FreqTransient
        if (select == 11)
        {
            for (int k = 0; k < samplerate; k++)
            {
                if (k < samplerate/2)
                {
                    signal[k] = Math.Sin(2*PI*52*k/samplerate);
                }
                if (k >= 3*samplerate/4 && k < (3*samplerate/4 + samplerate/16))
                {
                    signal[k] = Math.Sin(2*PI*23*k/samplerate);
                }
            }
        }

        // Sinus mit T <> ganzzahlig
        if (select == 12)
        {
            for (int k = 0; k < samplerate; k++)
            {
                signal[k] = Math.Sin(2 * PI * 12 * 0.95 * k / samplerate);
            }
        }

        // Sinus mit T = ganzzahlig
        if (select == 13)
        {
            for (int k = 0; k < 100; k++)
            {
                signal[k] = Math.Sin(2 * PI * 12 * k / 100);
            }
        }

        // Chirp-Signal
        if (select == 14)
        {
            double f = 0;

            for (int k = 0; k < samplerate; k++)
            {
                signal[k] = Math.Sin(2 * PI * f * k / samplerate);
                f += 1.0/8.0;
            }
        }

        // Walsh-Funktion 1. Ordnung
        if (select == 15)
        {
            for (int k = 0; k < samplerate; k++)
            {
                if (k < samplerate/2)
                {
                    signal[k] = 1;
                }
                else
                {
                    signal[k] = -1;
                }
            }
        }

        // Hanning-Fenster
        if (select == 16)
        {
            for (int k = 0; k < samplerate; k++)
            {
                signal[k] = 0.5 - 0.5 * Math.Cos(2*PI*k/samplerate);
            }
        }

        // Schalttransiente
        if (select == 17)
        {
            int tau = 0;

            for (int k = 0; k < samplerate; k++)
            {
                signal[k] = Math.Sin(2*PI*k/samplerate);

                if (k >= 1 * samplerate / 4)
                {
                    signal[k] += 0.9 * Math.Exp(-tau/30.0) * Math.Sin((1+PI/2.0) + 2*PI*57*k/samplerate);
                }

                tau += 1;
            }
        }

        // store argument data
        for (int k = 0; k < samplerate; k++)
        {
            arg[k] = k;// 2 * PI * k / samplerate;
        }
    }

    public void FFT(double[] x)
    {
        double tempr = 0;    // fr Tausch bei Bit-Umkehr
        double tempi = 0;    // fr Tausch bei Bit-Umkehr
        double wreal = 0;    // Drehfaktor (Realteil)
        double wimag = 0;    // Drehfaktor (Imaginrteil)
        double real1 = 0;    // Hilfsvariable
        double imag1 = 0;    // Hilfsvariable
        double real2 = 0;    // Hilfsvariable
        double imag2 = 0;    // Hilfsvariable
        int i = 0;
        int j = 0;
        int k = 0;
        int stages  = 0;    // 8 datapoints => 3 stages (8=2^3)
        int points  = 0;    // number of data points used in one stage
        int step    = 0;    // 
        int element = 0;

        // transform signal data into complex type
        for (k=0; k<samplerate; k++)
        {
            data[k].re = x[k];
            data[k].im = 0;
        }

		// Bit-Umkehr 
        for (j=0; j<samplerate-1; j++)
        {
            if (j < i)
            {
                tempr = data[j].re;
                tempi = data[j].im;
                data[j].re = data[i].re;
                data[j].im = data[i].im;
                data[i].re = tempr;
                data[i].im = tempi;
            }

            k = samplerate/2;
            while (k <= i)
            {
                i -= k;
                k /= 2;
            }

            i += k;
        }

        stages = (int)(Math.Log10((double)samplerate)/Math.Log10((double)2));
        points = 2;

        for (i=0; i<stages; i++)                            // 0..2 with N=8
        {
            // jede Iterationsstufe startet mit dem 1. Wert (Index 0)
            element = 0;

            // run over all butterflies of one stage (1..4, 1..2, 1 with N=8)
            for (j=1; j<=samplerate/points; j++)
            {
                step = points/2;

                for (k=0; k<step; k++)                    // 0, 0..1, 0..3 with N=8
                {
                    // 1. Zweig des Butterfly
                    wreal = Math.Cos(k*2.0*PI/points);
                    wimag = Math.Sin(k*2.0*PI/points);

                    real1 = data[element+k].re + 
                            (wreal*data[element+k+step].re - wimag*data[element+k+step].im);
                    imag1 = data[element+k].im + 
                            (wreal*data[element+k+step].im + wimag*data[element+k+step].re);

                    // 2. Zweig des Butterfly
                    wreal *= -1.0;
                    wimag *= -1.0;

                    real2 = data[element+k].re + 
                            (wreal*data[element+k+step].re - wimag*data[element+k+step].im);
                    imag2 = data[element+k].im + 
                            (wreal*data[element+k+step].im + wimag*data[element+k+step].re);

                    // Ergebnisse bernehmen
                    data[element+k].re      = real1;
                    data[element+k].im      = imag1;
                    data[element+k+step].re = real2;
                    data[element+k+step].im = imag2;
                }
                element += points;
            }
            points *= 2;                            // 2, 4, 8, 16, 32, 64, ...
        }

        // calculate amplitude and phase
        for (int f=0; f<samplerate; f++)
        {
            AS[f] = 2*Math.Sqrt(data[f].re*data[f].re + data[f].im*data[f].im)/samplerate;
            PS[f] = Math.Atan2(data[f].im, data[f].re + 1E-10);
        }
        AS[0] /= 2.0;   // Korrektur Gleichanteil
    }
}


class MainApp
{
    [STAThread]
    static void Main()
    {
        string userinput = "";
        int select = 0;

        while (select <= 17)
        {
            Spectrum mySpectrumObj = new Spectrum();

            Console.Write("Bitte Signal auswhlen:\n\n");
            Console.Write("1/1   Puls    ->  1\n");
            Console.Write("1/2   Puls    ->  2\n");
            Console.Write("1/4   Puls    ->  3\n");
            Console.Write("1/8   Puls    ->  4\n");
            Console.Write("1/16  Puls    ->  5\n");
            Console.Write("1/32  Puls    ->  6\n");
            Console.Write("1/64  Puls    ->  7\n");
            Console.Write("1/128 Puls    ->  8\n");
            Console.Write("1/8 Deltakamm ->  9\n");
            Console.Write("Spaltfunktion -> 10\n");
            Console.Write("FreqTransient -> 11\n");
            Console.Write("Sinus(T<>n)   -> 12\n");
            Console.Write("Sinus(T=n)    -> 13\n");
            Console.Write("Chirp         -> 14\n");
            Console.Write("Walsh 1.Ord.  -> 15\n");
            Console.Write("Hanning       -> 16\n");
            Console.Write("Transiente_2  -> 17\n");
            Console.WriteLine();
            userinput = Console.ReadLine();

            select = Convert.ToInt32(userinput);
            mySpectrumObj.CreateSignal(select);
            mySpectrumObj.FFT(mySpectrumObj.signal);

            // create a chart for displaying that data...
            DisplayChart mydisplay1 = new DisplayChart(720, 480, SeriesChartType.Line, "Time domain", "k", "x(k)");
            mydisplay1.AddData(mySpectrumObj.arg, mySpectrumObj.signal);

            DisplayChart mydisplay2 = new DisplayChart(720, 480, SeriesChartType.Column, "Frequency domain", "f", "|X(f)|");
            mydisplay2.AddData(mySpectrumObj.arg, mySpectrumObj.AS);

            DisplayChart mydisplay3 = new DisplayChart(720, 480, SeriesChartType.Line, "Frequency domain", "f", "Phi(f)");
            mydisplay3.AddData(mySpectrumObj.arg, mySpectrumObj.PS);

            // ...and run the chart-application
            Application.EnableVisualStyles();
            Application.Run(mydisplay1);
            Application.Run(mydisplay2);
            Application.Run(mydisplay3);

            mydisplay1.Dispose();
            mydisplay2.Dispose();
            mydisplay3.Dispose();
            Application.Exit();

            for (int i = 0; i < 20; i++) Console.WriteLine();
        }
    }
}
